#!/usr/bin/python
# filename: mongoimport.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import itertools
import multiprocessing as mp
import os
import subprocess
import sys
import time
import uuid

from abtools.pipeline import make_dir
from abtools.utils import progbar


def parse_arguments():
    import argparse
    parser = argparse.ArgumentParser("Performs the mongoimport operation on all files in the input directory.  Determines the appropriate collection using the filename.")
    parser.add_argument('-i', '--ip', dest='ip', default='localhost',
                        help="The IP address of the MongoDB server. Defaults to 'localhost'.")
    parser.add_argument('--port', dest='port', type=int, default=27017,
                        help="The MongoDB port. Defaults to 27017.")
    parser.add_argument('-u', '--user', dest='user', default=None,
                        help="Username for the MongoDB server. Not used if not provided.")
    parser.add_argument('-p', '--password', dest='password', default=None,
                        help="Password for the MongoDB server. Not used if not provided.")
    parser.add_argument('-f', '--in', dest='mongo_input_dir', required=True,
                        help="A directory containing multiple JSON files for import to MongoDB.")
    parser.add_argument('-d', '--db', dest='db', required=True,
                        help="The MongoDB database for import.")
    parser.add_argument('-l', '--log', dest='mongo_log', required=True,
                        help="Log file for the mongoimport stdout.")
    parser.add_argument('-t', '--temp', dest='temp', default=None,
                        help="Temp directory if splitting JSON files prior to mongoimport. \
                        if not provided, a temp directory will be created in the --in directory.")
    parser.add_argument('-e', '--delim1', dest='delim1', default=None,
                        help="The first character delimiter used to split the filename to get the collection name. \
                        If splitting with a single delimiter, use this option to provide the delimiter. Required.")
    parser.add_argument('--delim2', dest='delim2', default=None,
                        help="If splitting with two different delimiters, use this option to provide the second delimiter. \
                        Required if splitting with two delimiters.")
    parser.add_argument('-s', '--split1', dest='split1_pos', default=1, type=int,
                        help="Builds the collection name by truncating at the <split> occurance of the <delim> character. \
                        If splitting with multiple delimiters, this option is used to specifiy the occurance of the first delimiter. \
                        Default is 1.")
    parser.add_argument('--split2', dest='split2_pos', default=1, type=int,
                        help="If splitting with multiple delimiters, this option is used to specify the occurance of the \
                        second delimiter at which to split. \
                        Required if splitting with two different delimiters.")
    parser.add_argument('--split-file', dest='split_file', action='store_true', default=False,
                        help="Splits each input file into subfiles containing --split_file_lines lines. \
                        Useful when performing mongoimport via an SSH tunnel, where for some reason MongoDB \
                        errors when importing files greater than 16MB (even if no individual documents are >16MB).")
    parser.add_argument('--split=file-lines', dest='split_file_lines', type=int, default=500,
                        help="Number of lines in each mongoimport subfile (each line should be a complete JSON document).")
    parser.add_argument('-x', '--split-only', dest='split_only', default=False, action='store_true',
                        help="Instead of truncating the filename to get the collection name, takes only the split for the collection. \
                        Default is False.")
    return parser.parse_args()


class Args(object):
    """Holds arguments, mimics argparse's Namespace when running mongoimport as an imported module"""
    def __init__(self, ip='localhost', user=None, password=None, db=None,
                 mongo_input_dir=None, mongo_log=None, split_file=False, split_file_lines=500,
                 delim1=None, delim2=None,
                 split1_pos=1, split2_pos=1, split_only=False):
        super(Args, self).__init__()
        self.ip = ip
        self.user = user
        self.password = password
        self.mongo_input_dir = mongo_input_dir
        self.db = db
        self.mongo_log = mongo_log if mongo_log else sys.stdout
        self.split_file = split_file
        self.split_file_lines = int(split_file_lines)
        self.delim1 = delim1
        self.delim2 = delim2
        self.split1_pos = int(split1_pos)
        self.split2_pos = int(split2_pos)
        self.split_only = split_only


def mongo_import(json, db, coll, log, args):
    if args.split_file:
        jsons = split_file(json, args)
        multiprocess_mongoimport(jsons, db, coll, args)
    else:
        do_mongoimport(json, args.ip, args.port, db, coll, args.user, args.password)

    #     jsons = [json, ]
    # username = " -u {}".format(args.user) if args.user else ""
    # password = " -p {}".format(args.password) if args.password else ""
    # # user_password = "{}{} --authenticationDatabase admin".format(username, password)
    # for i, json in enumerate(jsons):
    #     mongo_cmd = "mongoimport --host {}:27017{}{} --db {} --collection {} --file {}".format(
    #         args.ip, username, password, db, coll, json)
    #     mongo = subprocess.Popen(mongo_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    #     mongo.communicate()
    #     if args.split_file:
    #         progbar.progress_bar(i + 1, len(jsons))
    # if args.split_file:
    #     print('')
    #     remove_temp_files(args)


def multiprocess_mongoimport(jsons, db, coll, args):
    progbar.progress_bar(0, len(jsons))
    async_results = []
    p = mp.Pool()
    for j in jsons:
        async_results.append(p.apply_async(do_mongoimport, args=(j, args.ip, args.port, db, coll, args.user, args.password)))
    monitor_results(async_results)
    remove_temp_files(args)
    print('')


def monitor_results(results):
    finished = 0
    jobs = len(results)
    while finished < jobs:
        time.sleep(1)
        ready = [ar for ar in results if ar.ready()]
        finished = len(ready)
        progbar.progress_bar(finished, jobs)
    progbar.progress_bar(finished, jobs)


def do_mongoimport(json, ip, port, db, coll, user, password):
    username = " -u {}".format(user) if user else ""
    password = " -p {}".format(password) if password else ""
    mongo_cmd = "mongoimport --host {}:{}{}{} --db {} --collection {} --file {}".format(
        ip, port, username, password, db, coll, json)
    mongo = subprocess.Popen(mongo_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    mongo.communicate()



def listdir_fullpath(d):
    if os.path.isdir(d):
        files = [os.path.join(d, f) for f in os.listdir(d) if f.split('.')[-1].lower() == 'json']
        if files:
            return sorted(files)
    else:
        if d.split('.')[-1].lower() == 'json':
            return [d, ]
    print("ERROR: no importable JSON files were found. Double-check your input file/directory")
    sys.exit(1)


def split_file(json, args):
    split_files = []
    temp_dir = args.temp if args.temp is not None else os.path.join(args.mongo_input_dir, 'temp')
    make_dir(temp_dir)
    with open(json) as f:
        for chunk in itertools.izip_longest(*[f] * args.split_file_lines):
            chunk = [c for c in chunk if c is not None]
            fname = os.path.join(temp_dir, str(uuid.uuid4()) + '.json')
            open(fname, 'w').write(''.join(chunk))
            split_files.append(fname)
    return split_files


def remove_temp_files(args):
    temp_dir = args.temp if args.temp is not None else os.path.join(args.mongo_input_dir, 'temp')
    files = listdir_fullpath(temp_dir)
    for f in files:
        os.unlink(f)



def get_collection(i, args):
    # split with two different delimiters
    if args.delim1 and args.delim2:
        bname = os.path.basename(i)
        split1 = args.delim1.join(bname.split(args.delim1)[args.split1_pos:])
        return args.delim2.join(split1.split(args.delim2)[:args.split2_pos])
    # split with a single delimiter
    elif args.delim1:
        delim = str(args.delim1)
        split = args.split1_pos
        if args.split_only:
            return os.path.basename(i).split(delim)[split - 1]
        if split <= 1:
            return os.path.basename(i).split(delim)[0]
        else:
            pre_collection = os.path.basename(i).split(delim)
            return delim.join(pre_collection[:split])
    else:
        return os.path.basename(i)


def preflight_checks(args):
    if not args.db:
        err = "ERROR: a MongoDB database name must be provided."
        raise RuntimeError(err)
    if not args.mongo_input_dir:
        err = "ERROR: a directory of JSON files must be provided."
        raise RuntimeError(err)
    if args.delim2 and not args.delim1:
        err = "ERROR: --delim1 must be provided with --delim2.\n"
        err += "If you only want to use one delimiter, use --delim1."
        raise RuntimeError(err)


def run(**kwargs):
    args = Args(**kwargs)
    main(args)


def main(args):
    preflight_checks(args)
    in_files = listdir_fullpath(args.mongo_input_dir)
    log_handle = open(args.mongo_log, 'a')
    open(args.mongo_log, 'w').write('')
    for i, f in enumerate(in_files):
        coll = get_collection(f, args)
        print "\n[ {} ] Importing {} into collection {}.".format(i + 1, os.path.basename(f), coll)
        log_handle.write('\n\n----------------------------------------')
        log_handle.write('File: {0}\nCollection: {1}'.format(f, coll))
        log_handle.write('----------------------------------------\n')
        mongo_import(f, args.db, coll, log_handle, args)
    print "\nDone. {0} files were imported into MongoDB.\n\n".format(len(in_files))


if __name__ == '__main__':
    args = parse_arguments()
    main(args)



#!/usr/bin/python
# filename: mutations.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import re
import traceback

from abtools import log


def nt_mutations(blast_result):
    logger = log.get_logger(__name__)
    try:
        return MutationsNT(blast_result)
    except:
        logger.debug('NT MUTATIONS ERROR: {}, {}\n'.format(blast_result.id,
                                                          blast_result.input_sequence))


def aa_mutations(blast_result):
    logger = log.get_logger(__name__)
    try:
        return MutationsAA(blast_result)
    except:
        logger.debug('AA MUTATIONS ERROR: {}, {}\n'.format(blast_result.id,
                                                          blast_result.input_sequence))



class MutationsNT(object):
    '''
    Structure for identifying and annotating nucleotide mutations.

    Input is a BlastResult object.
    '''
    def __init__(self, blast_result):
        self.all_mutations = self._find_all_mutations(blast_result)
        self.mutation_count = self._total_mutation_count(blast_result)
        self.germline_divergence = self._germline_divergence(blast_result)
        self.germline_identity = 100. - self.germline_divergence
        if blast_result.gene_type in ['variable', 'joining']:
            self.region_mutations = self._find_region_mutations(blast_result)
            self.region_mutation_count = self._region_mutation_count(blast_result)


    def _find_all_mutations(self, br):
        mutations = []
        q = br.query_alignment
        g = br.germline_alignment
        m = br.alignment_midline
        for i, pos in enumerate(m):
            if pos == '|':
                continue
            if q[i] == '-' or g[i] == '-':
                continue
            mutations.append({'pos': br.germline_start + i, 'mut': '{}>{}'.format(g[i], q[i])})
        return mutations


    def _total_mutation_count(self, br):
        return len(self.all_mutations)


    def _find_region_mutations(self, br):
        mutations = {}
        start_positions = self._region_start_positions(br)
        regions = self._region_list(br.gene_type)
        for r in regions:
            if br.regions.nt_seqs[r]:
                if br.gene_type == 'variable':
                    start = start_positions[r] if start_positions[r] > br.germline_start else br.germline_start
                if br.gene_type == 'joining':
                    start = br.germline_start + start_positions[r]
                mutations[r] = self._mutations_by_region(br, r, start)
            else:
                mutations[r] = []
        return mutations


    def _mutations_by_region(self, br, region, start):
        muts = []
        q = br.regions.nt_seqs[region]
        g = br.regions.germline_nt_seqs[region]
        for i, nt in enumerate(q):
            if nt == g[i] or nt == '-' or g[i] == '-':
                continue
            muts.append({'pos': i + start, 'mut': '{}>{}'.format(g[i], q[i])})
        return muts


    def _region_mutation_count(self, br):
        mutation_counts = {}
        for r in self._region_list(br.gene_type):
            if br.regions.nt_seqs[r]:
                mutation_counts[r] = len(self.region_mutations[r])
            else:
                mutation_counts[r] = 0
        return mutation_counts


    def _region_list(self, gene_type):
        if gene_type == 'variable':
            return ['FR1', 'CDR1', 'FR2', 'CDR2', 'FR3']
        if gene_type == 'joining':
            return ['FR4', ]


    def _region_start_positions(self, br):
        positions = {}
        if br.gene_type == 'variable':
            positions['FR1'] = 0
            for i, region in enumerate(self._region_list(br.gene_type)[1:]):
                positions[region] = br.regions.raw_positions[i]
        if br.gene_type == 'joining':
            for i, region in enumerate(self._region_list(br.gene_type)):
                positions[region] = br.regions.raw_positions[i]
        return positions


    def _germline_divergence(self, br):
        return 100. * self.mutation_count / len(br.query_alignment.replace('-', ''))



class MutationsAA(object):
    '''
    Structure for identifying and annotating amino acid mutations.

    Input is a BlastResult object.
    '''
    def __init__(self, blast_result):
        self.region_mutations = self._find_region_mutations(blast_result)
        self.region_mutation_count = self._region_mutation_count(blast_result)
        self.all_mutations = self._find_all_mutations(blast_result)
        self.mutation_count = self._total_mutation_count(blast_result)
        self.germline_divergence = self._germline_divergence(blast_result)
        self.germline_identity = 100. - self.germline_divergence


    def _find_region_mutations(self, br):
        mutations = {}
        start_positions = self._region_start_positions(br)
        regions = self._region_list(br.gene_type)
        for r in regions:
            if br.regions.aa_seqs[r]:
                if br.gene_type == 'variable':
                    start = start_positions[r] / 3 if start_positions[r] > br.germline_start else br.germline_start / 3
                if br.gene_type == 'joining':
                    start = (br.germline_start + start_positions[r]) / 3
                mutations[r] = self._mutations_by_region(br, r, start)
            else:
                mutations[r] = []
        return mutations


    def _mutations_by_region(self, br, region, start):
        muts = []
        q = br.regions.aa_seqs[region]
        g = br.regions.germline_aa_seqs[region]
        for i, aa in enumerate(q):
            if aa == g[i] or aa == '-' or g[i] == '-':
                continue
            muts.append({'pos': i + start, 'mut': '{}>{}'.format(g[i], q[i])})
        return muts


    def _region_mutation_count(self, br):
        mutation_counts = {}
        for r in self._region_list(br.gene_type):
            if br.regions.aa_seqs[r]:
                mutation_counts[r] = len(self.region_mutations[r])
            else:
                mutation_counts[r] = 0
        return mutation_counts


    def _region_list(self, gene_type):
        if gene_type == 'variable':
            return ['FR1', 'CDR1', 'FR2', 'CDR2', 'FR3']
        if gene_type == 'joining':
            return ['FR4', ]


    def _region_start_positions(self, br):
        positions = {}
        if br.gene_type == 'variable':
            positions['FR1'] = 0
            for i, region in enumerate(self._region_list(br.gene_type)[1:]):
                positions[region] = br.regions.raw_positions[i] / 3
        if br.gene_type == 'joining':
            for i, region in enumerate(self._region_list(br.gene_type)):
                positions[region] = br.regions.raw_positions[i] / 3
        return positions


    def _find_all_mutations(self, br):
        mutations = []
        for r in self._region_list(br.gene_type):
            if r in self.region_mutations:
                mutations.extend(self.region_mutations[r])
        return mutations


    def _total_mutation_count(self, br):
        return len(self.all_mutations)


    def _germline_divergence(self, br):
        aa_length = sum([len(aa.replace('-', '')) for aa in br.regions.aa_seqs.values() if aa])
        if aa_length:
            return 100. * self.mutation_count / aa_length
        return 0.0



"""
@package ssw_wrap
@brief Simple python wrapper for SSW align library
To use the dynamic library libssw.so you may need to modify the LD_LIBRARY_PATH environment
variable to include the library directory (export LD_LIBRARY_PATH=$PWD) or for definitive
inclusion of the lib edit /etc/ld.so.conf and add the path or the directory containing the
library and update the cache by using /sbin/ldconfig as root
@copyright  [The MIT licence](http://opensource.org/licenses/MIT)
@author     Clement & Adrien Leger - 2014
"""

#~~~~~~~GLOBAL IMPORTS~~~~~~~#
# Standard library packages
from ctypes import *

import os
import platform


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
class CAlignRes(Structure):
    """
    @class  SSWAlignRes
    @brief  ctypes Structure with s_align struct mapping returned by SSWAligner.Align func
            Correspond to the structure of the query profile
    """
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

    #~~~~~~~Ctype Structure~~~~~~~#
    _fields_ = [('score', c_uint16),
                ('score2', c_uint16),
                ('ref_begin', c_int32),
                ('ref_end', c_int32),
                ('query_begin', c_int32),
                ('query_end', c_int32),
                ('ref_end2', c_int32),
                ('cigar', POINTER(c_uint32)),
                ('cigarLen', c_int32)]

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
class Aligner(object):
    """
    @class  SSWAligner
    @brief Wrapper for SSW align library
    """
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

    #~~~~~~~CLASS VARIABLES~~~~~~~#

    # Dictionnary to map Nucleotide to int as expected by the SSW C library
    base_to_int = { 'A':0, 'C':1, 'G':2, 'T':3, 'N':4, 'a':0, 'c':1, 'g':2, 't':3, 'n':4}
    int_to_base = { 0:'A', 1:'C', 2:'G', 3:'T', 4:'N'}

    # Load the ssw library using ctypes
    _mod_dir = os.path.dirname(os.path.abspath(__file__))
    _ssw_path = os.path.join(_mod_dir, 'libssw_{}.so'.format(platform.system().lower()))
    libssw = cdll.LoadLibrary(_ssw_path)

    # Init and setup the functions pointer to map the one specified in the SSW lib
    # ssw_init method
    ssw_init = libssw.ssw_init
    ssw_init.restype = c_void_p
    ssw_init.argtypes = [POINTER(c_int8), c_int32, POINTER(c_int8), c_int32, c_int8]
    # init_destroy function
    init_destroy = libssw.init_destroy
    init_destroy.restype = None
    init_destroy.argtypes =  [c_void_p]
    # ssw_align function
    ssw_align = libssw.ssw_align
    ssw_align.restype = POINTER(CAlignRes)
    ssw_align.argtypes = [c_void_p, POINTER(c_int8), c_int32, c_uint8, c_uint8, c_uint8, c_uint16, c_int32, c_int32]
    # align_destroy function
    align_destroy = libssw.align_destroy
    align_destroy.restype = None
    align_destroy.argtypes = [POINTER(CAlignRes)]

    #~~~~~~~FONDAMENTAL METHODS~~~~~~~#

    def __repr__(self):
        msg = self.__str__()
        msg += "SCORE PARAMETERS:\n"
        msg += " Gap Weight     Open: {}     Extension: {}\n".format(-self.gap_open, -self.gap_extend)
        msg += " Align Weight   Match: {}    Mismatch: {}\n\n".format(self.match, -self.mismatch)
        msg += " Match/mismatch Score matrix\n"
        msg += " \tA\tC\tG\tT\tN\n"
        msg += " A\t{}\t{}\t{}\t{}\t{}\n".format(self.match, -self.mismatch, -self.mismatch, -self.mismatch, 0)
        msg += " C\t{}\t{}\t{}\t{}\t{}\n".format(-self.mismatch, self.match, -self.mismatch, -self.mismatch, 0)
        msg += " G\t{}\t{}\t{}\t{}\t{}\n".format(-self.mismatch, -self.mismatch, self.match, -self.mismatch, 0)
        msg += " T\t{}\t{}\t{}\t{}\t{}\n".format(-self.mismatch, -self.mismatch, -self.mismatch, self.match, 0)
        msg += " N\t{}\t{}\t{}\t{}\t{}\n\n".format(0,0,0,0,0)
        msg += "RESULT PARAMETERS:\n"
        msg += " Report cigar           {}\n".format(self.report_cigar)
        msg += " Report secondary match {}\n\n".format(self.report_secondary)
        msg += "REFERENCE SEQUENCE :\n"
        if self.ref_len <= 50:
            msg += "".join([self.int_to_base[i] for i in self.ref_seq])+"\n"
        else:
            msg += "".join([self.int_to_base[self.ref_seq[i]] for i in range(50)])+"...\n"
        msg += " Lenght :{} nucleotides\n".format(self.ref_len)
        return msg

    def __str__(self):
        return "\n<Instance of {} from {} >\n".format(self.__class__.__name__, self.__module__)

    def __init__(self,
                ref_seq="",
                match=2,
                mismatch=2,
                gap_open=3,
                gap_extend=1,
                report_secondary=False,
                report_cigar=False):
        """
        Initialize object by creating an interface with ssw library fonctions
        A reference sequence is also assigned to the object for multiple alignment against queries
        with the align function
        @param ref_seq Reference sequence as a python string (case insensitive)
        @param match Weight for a match
        @param mismatch Absolute value of mismatch penalty
        @param gap_open Absolute value of gap open penalty
        @param gap_extend Absolute value of gap extend penalty
        @param report_secondary Report the 2nd best alignement if true
        @param report_cigar Report cigar string if true
        """

        # Store overall alignment parameters
        self.report_secondary = report_secondary
        self.report_cigar = report_cigar

        # Set gap penalties
        self.set_gap(gap_open, gap_extend)

        # Set the cost matrix
        self.set_mat(match, mismatch)

        # Set the reference sequence
        self.set_ref(ref_seq)

    #~~~~~~~SETTERS METHODS~~~~~~~#

    def set_gap(self, gap_open=3, gap_extend=1):
        """
        Store gapopen and gap extension penalties
        """
        self.gap_open = gap_open
        self.gap_extend = gap_extend


    def set_mat(self, match=2, mismatch=2):
        """
        Store match and mismatch scores then initialize a Cost matrix and fill it with match and
        mismatch values. Ambiguous base: no penalty
        """
        self.match = match
        self.mismatch = mismatch

        mat_decl = c_int8 * 25
        self.mat = mat_decl(match, -mismatch, -mismatch, -mismatch, 0,
                            -mismatch, match, -mismatch, -mismatch, 0,
                            -mismatch, -mismatch, match, -mismatch, 0,
                            -mismatch, -mismatch, -mismatch, match, 0,
                            0, 0, 0, 0, 0)

    def set_ref(self, ref_seq):
        """
        Determine the size of the ref sequence and cast it in a c type integer matrix
        """
        if ref_seq:
            self.ref_len = len(ref_seq)
            self.ref_seq = self._DNA_to_int_mat (ref_seq, self.ref_len)
        else:
            self.ref_len = 0
            self.ref_seq = ""

    #~~~~~~~PUBLIC METHODS~~~~~~~#

    def align(self, query_seq, min_score=0, min_len=0):
        """
        Perform the alignment of query against the object reference sequence
        @param query_seq Query sequence as a python string (case insensitive)
        @param min_score Minimal score of match. None will be return in case of filtering out
        @param min_len Minimal length of match. None will be return in case of filtering out
        @return A SSWAlignRes Object containing informations about the alignment.
        """
        # Determine the size of the ref sequence and cast it in a c type integer matrix
        query_len = len(query_seq)
        query_seq = self._DNA_to_int_mat (query_seq, query_len)

        # Create the query profile using the query sequence
        profile = self.ssw_init(query_seq, # Query seq in c type integers
                                c_int32(query_len), # Length of Queryseq in bites
                                self.mat, # Score matrix
                                5, # Square root of the number of elements in mat
                                2) # flag = no estimation of the best alignment score

        # Setup the mask_len parameters = distance between the optimal and suboptimal alignment
        # if < 15, the function will NOT return the suboptimal alignment information

        if query_len > 30:
            mask_len = query_len/2
        else:
            mask_len = 15

        c_result = self.ssw_align (profile, # Query profile
                                self.ref_seq, # Ref seq in c type integers
                                c_int32(self.ref_len), # Length of Refseq in bites
                                self.gap_open, # Absolute value of gap open penalty
                                self.gap_extend, # absolute value of gap extend penalty
                                1, # Bitwise FLAG for output values = return all
                                0, # Score filter = return all
                                0, # Distance filter = return all
                                mask_len) # Distance between the optimal and suboptimal alignment

        # Transform the Cstructure into a python object if score and lenght match the requirements
        score = c_result.contents.score
        match_len  = c_result.contents.query_end - c_result.contents.query_begin + 1

        if score >= min_score and match_len >= min_len:
            py_result = PyAlignRes(c_result, query_len, self.report_secondary, self.report_cigar)
        else:
            py_result = None

        # Free reserved space by ssw.init and ssw_init methods.
        self._init_destroy(profile)
        self._align_destroy(c_result)

        # Return the object
        return py_result

    #~~~~~~~PRIVATE METHODS~~~~~~~#

    def _DNA_to_int_mat (self, seq, len_seq):
        """
        Cast a python DNA string into a Ctype int8 matrix
        """
        # Declare the matrix
        query_num_decl = c_int8 * len_seq
        query_num = query_num_decl()

        # for each letters in ATCGN transform in integers thanks to self.base_to_int
        for i in range(len_seq):
            try:
                value = self.base_to_int[seq[i]]
            # if the base is not in the canonic DNA bases assign 4 as for N
            except KeyError:
                value = 4
            finally:
                query_num[i] = value

        return query_num

    def _init_destroy(self, profile):
        """
        Free the space alocated for the matrix used by init
        """
        self.init_destroy(profile)

    def _align_destroy(self, align):
        """
        Free the space alocated for the matrix used by align
        """
        self.align_destroy(align)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
class PyAlignRes(object):
    """
    @class  PyAlignRes
    @brief  Extract and verify result from a CAlignRes structure. A comprehensive python
    object is created according to user requirements (+- cigar string and secondary alignment)
    """
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#

    #~~~~~~~CLASS VARIABLES~~~~~~~#

    # Load the ssw library using ctypes
    _mod_dir = os.path.dirname(os.path.abspath(__file__))
    _ssw_path = os.path.join(_mod_dir, 'libssw_{}.so'.format(platform.system().lower()))
    libssw = cdll.LoadLibrary(_ssw_path)
    # libssw = cdll.LoadLibrary('ssw/libssw_{}.so'.format(platform.system().lower()))
    # libssw = cdll.LoadLibrary('libssw.so')

    # Init and setup the functions pointer to map the one specified in the SSW lib
    # cigar_int_to_len function
    cigar_int_to_len = libssw.cigar_int_to_len
    cigar_int_to_len.restype = c_int32
    cigar_int_to_len.argtypes = [c_int32]
    # cigar_int_to_op function
    cigar_int_to_op = libssw.cigar_int_to_op
    cigar_int_to_op.restype = c_char
    cigar_int_to_op.argtypes = [c_int32]

    #~~~~~~~FONDAMENTAL METHOD~~~~~~~#

    def __repr__(self):
        msg = self.__str__()
        msg += "OPTIMAL MATCH\n"
        msg += "Score            {}\n".format(self.score)
        msg += "Reference begin  {}\n".format(self.ref_begin)
        msg += "Reference end    {}\n".format(self.ref_end)
        msg += "Query begin      {}\n".format(self.query_begin)
        msg += "Query end        {}\n".format(self.query_end)

        if self.cigar_string:
            msg += "Cigar_string     {}\n".format(self.cigar_string)

        if self.score2:
            msg += "SUB-OPTIMAL MATCH\n"
            msg += "Score 2           {}\n".format(self.score2)
            msg += "Ref_end2          {}\n".format(self.ref_end2)

        return msg

    def __str__(self):
        return "\n<Instance of {} from {} >\n".format(self.__class__.__name__, self.__module__)


    def __init__ (self, Res, query_len, report_secondary=False, report_cigar=False):
        """
        Parse CAlignRes structure and copy its values in object variables
        @param Res A CAlignRes structure
        @param query_len length of the query sequence
        @param report_secondary Report the 2nd best alignement if true
        @param report_cigar Report cigar string if true
        """
        # Parse value in the C type structure pointer
        # Minimal mandatory parameters
        self.score = Res.contents.score
        self.ref_begin = Res.contents.ref_begin
        self.ref_end = Res.contents.ref_end
        self.query_begin = Res.contents.query_begin
        self.query_end = Res.contents.query_end

        # Information for sub-optimal match if require and available
        score2 = Res.contents.score2
        if report_secondary and score2 != 0:
            self.score2 = score2
            self.ref_end2 = Res.contents.ref_end2
        else:
            self.score2 = None
            self.ref_end2 = None

        # Cigar Information if CIGAR string if require and available
        cigar_len = Res.contents.cigarLen
        if report_cigar and cigar_len > 0:
            self.cigar_string = self._cigar_string (Res.contents.cigar, cigar_len, query_len)
        else:
            self.cigar_string = None

    #~~~~~~~PRIVATE METHODS~~~~~~~#

    def _cigar_string(self, cigar, cigar_len, query_len):
        """
        Convert cigar and cigarLen into an human readable Cigar string as in SAM files
        """
        # Empty string for iterative writing of the cigar string
        cigar_string = ""

        # If the query match do not start at its first base
        # = introduce a softclip at the begining
        if self.query_begin > 0:
            op_len = self.query_begin
            op_char = "S"
            cigar_string += '{}{}'.format(op_len, op_char)

        # Iterate over the cigar (pointer to a vector of int)
        for i in range(cigar_len):
            op_len = self.cigar_int_to_len(cigar[i])
            op_char = self.cigar_int_to_op(cigar[i])
            cigar_string += '{}{}'.format(op_len, op_char)

        # If the lenght of bases aligned is shorter than the overall query length
        # = introduce a softclip at the end
        end_len = query_len - self.query_end - 1
        if  end_len != 0:
            op_len = end_len
            op_char = "S"
            cigar_string += '{}{}'.format(op_len, op_char)

        return cigar_string



#!/usr/bin/python
# filename: basespace.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function, absolute_import

import json
import os
import platform
import sys
import time

from BaseSpacePy.api.BaseSpaceAPI import BaseSpaceAPI
from BaseSpacePy.model.QueryParameters import QueryParameters as qp

from abtools import log
logger = log.get_logger('basespace')


class BaseSpace(object):
    """docstring for BaseSpace"""
    def __init__(self, project_id=None, project_name=None, undetermined=False):
        super(BaseSpace, self).__init__()
        # self.log = log
        # BaseSpace credentials
        creds = self._get_credentials()
        self.client_key = creds['client_id']
        self.client_secret = creds['client_secret']
        self.access_token = creds['access_token']
        self.version = creds['version']
        self.api_server = creds['api_server']
        self.api = BaseSpaceAPI(self.client_key, self.client_secret, self.api_server, self.version, AccessToken=self.access_token)
        self.params = qp(pars={'Limit': 1024, 'SortDir': 'Desc'})
        if project_id is not None:
            self.project_id = project_id
            self.project_name = None
        elif project_name is not None:
            self.project_name = project_name
            self.project_id = self._get_project_id_from_name(project_name)
        else:
            self.project_id, self.project_name = self._user_selected_project_id()


    def _get_credentials(self):
        # BaseSpace credentials file should be in JSON format
        cred_file = os.path.expanduser('~/.abstar/basespace_credentials')
        cred_handle = open(cred_file, 'r')
        return json.load(cred_handle)


    def _get_project_id_from_name(self):
        projects = self.api.getProjectByUser(queryPars=self.params)
        for project in projects:
            name = project.Name.encode('ascii', 'ignore')
            if name == self.project_name:
                return project.Id
        print('No projects matched the given project name ({})'.format(name))
        sys.exit(1)


    def _user_selected_project_id(self):
        projects = self.api.getProjectByUser(queryPars=self.params)
        self.print_basespace_project()
        offset = 0
        while True:
            for i, project in enumerate(projects[offset * 25:(offset * 25) + 25]):
                project_name = project.Name.encode('ascii', 'ignore')
                print('[ {} ] {}'.format(i + (offset * 25), project_name))
            print('')
            project_index = raw_input("Select the project number (or 'next' to see more projects): ")
            try:
                project_index = int(project_index)
                return projects[project_index].Id, projects[project_index].Name.encode('ascii', 'ignore')
            except:
                offset += 1
        return projects[project_index].Id, projects[project_index].Name.encode('ascii', 'ignore')


    def _get_projects(self, start=0):
        projects = self.api.getProjectByUser(queryPars=self.params)
        self.print_basespace_project()
        for i, project in enumerate(projects[:25]):
            project_name = project.Name.encode('ascii', 'ignore')
            print('[ {} ] {}'.format(i, project_name))
        print('')
        return projects


    def _get_samples(self, project_id):
        samples = []
        offset = 0
        while True:
            query_params = qp(pars={'Limit': 1024, 'SortDir': 'Asc', 'Offset': offset * 1024})
            s = self.api.getSamplesByProject(self.project_id, queryPars=query_params)
            if not s:
                break
            samples.extend(s)
            offset += 1
        return samples


    def _get_files(self):
        files = []
        samples = self._get_samples(self.project_id)
        for sample in samples:
            files.extend(self.api.getFilesBySample(sample.Id, queryPars=self.params))
        return files


    def download(self, direc):
        files = self._get_files()
        self.print_download_info(files)
        start = time.time()
        for i, f in enumerate(files):
            # self.log.write('[ {} ] {}\n'.format(i, str(f)))
            logger.info('[ {} ] {}'.format(i, str(f)))
            f.downloadFile(self.api, direc)
        end = time.time()
        self.print_completed_download_info(start, end)
        return len(files)


    def print_basespace_project(self):
        print('')
        print('')
        print('========================================')
        print('BaseSpace Project Selection')
        print('========================================')
        print('')


    def print_download_info(self, files):
        logger.info('')
        logger.info('')
        logger.info('========================================')
        logger.info('Downloading files from BaseSpace')
        logger.info('========================================')
        logger.info('')
        logger.info('Identified {0} files for download.'.format(len(files)))
        logger.info('')


    def print_completed_download_info(self, start, end):
        logger.info('')
        logger.info('Download completed in {0} seconds'.format(end - start))


# def _setup_logging():
#     try:
#         from abtools.utils import log
#         global logger
#         logger = log.get_logger('basespace')
#     except ImportError:
#         import logging
#         fmt = '[%(levelname)s] %(name)s %(asctime)s %(message)s'
#         logging.basicConfig()
#         global logger
#         logger = logging.getLogger('basespace')
#         formatter = logging.Formatter("%(message)s")
#         ch = logging.StreamHandler()
#         ch.setFormatter(formatter)
#         ch.setLevel(logging.INFO)
#         logger.addHandler(ch)


def download(direc, project_id=None, project_name=None, undetermined=False):
    # _setup_logging()
    bs = BaseSpace(project_id, project_name, undetermined)
    return bs.download(direc)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser("Downloads sequencing data from BaseSpace, Illumina's cloud storage environment.")
    parser.add_argument('-i', '--in',
                        dest='input',
                        required=True,
                        help="The input file, to be split and processed in parallel. \
                        If a directory is given, all files in the directory will be iteratively processed.")
    args = parser.parse_args()
    download(args.input)



# import alignment
# import basespace
# import indels
# import isotype
# import junction
# import mongoimport
# import mutations
# import output
# import pandaseq
# import productivity
# import regions
# import sequence
# import vdj



#!/usr/bin/python
# filename: codons.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import logging
import traceback

from abtools import log


def parse_codons(vdj, gapped):
    logger = log.get_logger(__name__)
    try:
        codons = Codons(vdj, gapped)
        return codons
    except:
        logger.debug('PARSE CODONS ERROR: {}, {}'.format(vdj.id, vdj.raw_input))
        logger.debug(traceback.format_exc())


class Codons(object):
    """docstring for Codons"""
    def __init__(self, vdj, gapped=False):
        super(Codons, self).__init__()
        self.gapped = gapped
        self.vdj_codons = self._get_vdj_codons(vdj)
        self.vdj_germ_codons = self._get_vdj_codons(vdj)
        self.vdj_codon_regions = self._get_vdj_codon_regions(vdj)
        self.v_codons = self._get_v_codons(vdj)
        self.v_germ_codons = self._get_v_germ_codons(vdj)
        self.j_codons = self._get_j_codons(vdj)
        self.j_germ_codons = self._get_j_germ_codons(vdj)


    def _get_vdj_codons(self, vdj):
        offset = (vdj.query_reading_frame * 2) % 3
        if self.gapped:
            trunc_vdj = vdj.gapped_vdj_nt[offset:]
        else:
            trunc_vdj = vdj.vdj_nt[offset:]
        codons = [trunc_vdj[i:i + 3] for i in range(0, len(trunc_vdj), 3)]
        if len(codons[-1]) != 3:
            codons = codons[:-1]
        return codons

    def _get_vdj_codons(self, vdj):
        offset = (vdj.query_reading_frame * 2) % 3
        if self.gapped:
            trunc_vdj = vdj.gapped_vdj_germ_nt[offset:]
        else:
            trunc_vdj = vdj.vdj_germ_nt[offset:]
        codons = [trunc_vdj[i:i + 3] for i in range(0, len(trunc_vdj), 3)]
        if len(codons[-1]) != 3:
            codons = codons[:-1]
        return codons

    def _get_vdj_codon_regions(self, vdj):
        offset = (vdj.query_reading_frame * 2) % 3
        if self.gapped:
            trunc_region = vdj.gapped_vdj_region_string[offset:]
        else:
            trunc_region = vdj.vdj_region_string[offset:]
        codons = [trunc_region[i:i + 3] for i in range(0, len(trunc_region), 3)]
        if len(codons[-1]) != 3:
            codons = codons[:-1]
        return codons

    def _get_v_codons(self, vdj):
        vbr = vdj.v
        offset = (vdj.query_reading_frame * 2) % 3
        trunc_v = vbr.query_alignment[offset:]
        if not self.gapped:
            trunc_v.replace('-', '')
        codons = [trunc_v[i:i + 3] for i in range(0, len(trunc_v), 3)]
        if len(codons[-1]) != 3:
            codons = codons[:-1]
        return codons

    def _get_v_germ_codons(self, vdj):
        vbr = vdj.v
        offset = (vdj.query_reading_frame * 2) % 3
        trunc_v = vbr.germline_alignment[offset:]
        if not self.gapped:
            trunc_v.replace('-', '')
        codons = [trunc_v[i:i + 3] for i in range(0, len(trunc_v), 3)]
        if len(codons[-1]) != 3:
            codons = codons[:-1]
        return codons

    def _get_j_codons(self, vdj):
        jbr = vdj.j
        rf = jbr.germline_start % 3
        trunc_j = jbr.query_alignment[rf:]
        codons = [trunc_j[i:i + 3] for i in range(0, len(trunc_j), 3)]
        if len(codons[-1]) != 3:
            codons = codons[:-1]
        return codons

    def _get_j_germ_codons(self, vdj):
        jbr = vdj.j
        rf = jbr.germline_start % 3
        trunc_j = jbr.germline_alignment[rf:]
        codons = [trunc_j[i:i + 3] for i in range(0, len(trunc_j), 3)]
        if len(codons[-1]) != 3:
            codons = codons[:-1]
        return codons



#!/usr/bin/python
# filename: junction.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import traceback

from abtools import log
from abtools.alignment import global_alignment, local_alignment


def get_junction(vdj, germ=False):
    logger = log.get_logger(__name__)
    try:
        return Junction(vdj, germ)
    except Exception, err:
        logger.debug('JUNCTION ERROR: {id}'.format(id=vdj.id))
        logger.debug(traceback.format_exc())


class Junction(object):
    """docstring for Junction"""
    def __init__(self, vdj, germ):
        self.is_germ = germ
        self.junction_aa_start_pos = self._find_junction_aa_start(vdj)
        self.junction_aa_end_pos = self._find_junction_aa_end(vdj)
        self.junction_aa = self._get_junction_aa_sequence(vdj)
        self.junction_nt_start_pos = self._find_junction_nt_start(vdj)
        self.junction_nt_end_pos = self._find_junction_nt_end(vdj)
        self.junction_nt = self._get_junction_nt_sequence(vdj)
        self.cdr3_aa = self.junction_aa[1:-1]
        self.cdr3_nt = self.junction_nt[3:-3]
        if vdj.d:
            self.n1_nt = vdj.vdj_nt[vdj.n1_start:vdj.n1_end]
            self.d_nt = vdj.vdj_nt[vdj.d_start:vdj.d_end]
            self.n2_nt = vdj.vdj_nt[vdj.n2_start:vdj.n2_end]
            self.n_nt = None
            self.d_start_position_nt = self._get_d_start_position_nt(vdj)
            self.d_end_position_nt = self._get_d_end_position(vdj)
            self.d_dist_from_cdr3_start_nt = self.d_start_position_nt
            self.d_dist_from_cdr3_end_nt = self._get_d_dist_from_cdr3_end_nt(vdj)
        else:
            self.n1_nt = None
            self.d_nt = None
            self.n2_nt = None
            self.n_nt = vdj.vdj_nt[vdj.n_start:vdj.n_end]


    def _find_junction_aa_start(self, vdj):
        fr3_start = sum(vdj.v.regions.aa_lengths.values()) - vdj.v.regions.aa_lengths['FR3']
        alignments = []
        germ = self._get_v_chunk(vdj)
        for i in range(fr3_start, len(vdj.vdj_aa) - 4):
            chunk = vdj.vdj_aa[i:i + 6]
            alignments.append(global_alignment((str(i), chunk), ('germ', germ),
                                               aa=True, matrix='blosum62',
                                               gap_open=-100, gap_extend=-50))
        alignments.sort(key=lambda x: x.score, reverse=True)
        top = alignments[0]
        if vdj.vdj_aa[int(top.query.id) + 4] != 'C':
            start_pos = self._verify_junction_aa_start(vdj, alignments)
        else:
            start_pos = int(top.query.id) + 4
        return start_pos


    def _verify_junction_aa_start(self, vdj, alignments):
        if alignments[1].score / alignments[0].score >= 0.75:
            if int(alignments[1].query.id) < int(alignments[0].query.id) and vdj.vdj_aa[int(alignments[1].query.id) + 4] == 'C':
                return int(alignments[1].query.id) + 4
        return int(alignments[0].query.id) + 4


    def _find_junction_aa_end(self, vdj):
        alignments = []
        germ = self._get_j_chunk(vdj)
        for i in range(self.junction_aa_start_pos, len(vdj.vdj_aa) - 4):
            chunk = (i, vdj.vdj_aa[i:i + 5])
            alignments.append(global_alignment(chunk, germ, aa=True, matrix='blosum62'))
        alignments.sort(key=lambda x: x.score, reverse=True)
        return int(alignments[0].query.id) + 1


    def _get_junction_aa_sequence(self, vdj):
        start = self.junction_aa_start_pos
        end = self.junction_aa_end_pos
        if self.is_germ:
            return vdj.vdj_germ_aa[start:end]
        return vdj.vdj_aa[start:end]


    def _find_junction_nt_start(self, vdj):
        return self.junction_aa_start_pos * 3 + (vdj.query_reading_frame * 2) % 3


    def _find_junction_nt_end(self, vdj):
        return self.junction_nt_start_pos + len(self.junction_aa) * 3


    def _get_junction_nt_sequence(self, vdj):
        start = self.junction_nt_start_pos
        end = self.junction_nt_end_pos
        if self.is_germ:
            return vdj.vdj_germ_nt[start:end]
        return vdj.vdj_nt[start:end]


    def _get_d_start_position_nt(self, vdj):
        a = local_alignment(self.d_nt, self.cdr3_nt,
                            gap_open_penalty=22, gap_extend_penalty=1)
        d_start = a.target_begin
        return d_start


    def _get_d_end_position(self, vdj):
        return self.d_start_position_nt + len(vdj.d.sequence)


    def _get_d_dist_from_cdr3_end_nt(self, vdj):
        return len(self.cdr3_nt) - (self.d_dist_from_cdr3_start_nt + len(self.d_nt))


    def _get_v_chunk(self, vdj):
        trunc_gene = vdj.v.top_germline[:5]
        germs = {
            'human':    {    'IGHV1': 'AVYYCA',
                            'IGHV2': 'ATYYCA',
                            'IGHV3': 'AVYYCA',
                            'IGHV4': 'AVYYCA',
                            'IGHV5': 'AMYYCA',
                            'IGHV6': 'AVYYCA',
                            'IGHV7': 'AVYYCA',
                            'IGKV1': 'ATYYCQ',
                            'IGKV2': 'GVYYCM',
                            'IGKV3': 'AVYYCQ',
                            'IGKV4': 'AVYYCQ',
                            'IGLV1': 'ADYYCQ',
                            'IGLV2': 'ADYYCS',
                            'IGLV3': 'ADYYCQ',
                            'IGLV4': 'ADYYCQ',
                            'IGLV5': 'ADYYCM',
                            'IGLV6': 'ADYYCQ',
                            'IGLV7': 'AEYYCL',
                            'IGLV8': 'SDYYCV',
                            'IGLV9': 'SDYHCG'},

            'macaque':    {    'IGHV1': 'AVYYCA',
                            'IGHV2': 'ATYYCA',
                            'IGHV3': 'AVYYCA',
                            'IGHV4': 'AVYYCA',
                            'IGHV5': 'ATYYCA',
                            'IGHV6': 'AVYYCA',
                            'IGHV7': 'AVYYCA',
                            'IGKV1': 'ATYYCQ',
                            'IGKV2': 'GVYYCM',
                            'IGKV3': 'AVYYCQ',
                            'IGKV4': 'AVYYCQ',
                            'IGKV5': 'AYYFCQ',
                            'IGKV6': 'ATYYCQ',
                            'IGKV7': 'ADYYCL',
                            'IGLV1': 'ADYYCQ',
                            'IGLV2': 'ADYYCS',
                            'IGLV3': 'ADYYCQ',
                            'IGLV4': 'ADYYCQ',
                            'IGLV5': 'ADYYCM',
                            'IGLV6': 'ADYYCQ',
                            'IGLV7': 'AEYYCW',
                            'IGLV8': 'SDYYCT',
                            'IGLV9': 'SDYHCG'
                            },
            
            'mouse':    {    'IGHV1': 'AVYYCA',
                            'IGHV2': 'AIYYCA',
                            'IGHV3': 'ATYYCA',
                            'IGHV4': 'ALYYCA',
                            'IGHV5': 'AMYYCA',
                            'IGHV6': 'GIYYCT',
                            'IGHV7': 'ATYYCA',
                            'IGHV8': 'ATYYCA',
                            'IGHV9': 'ATYFCA',
                            'IGKV1': 'ATYYCQ',
                            'IGKV2': 'GVYYCA',
                            'IGKV3': 'ATYYCQ',
                            'IGKV4': 'ATYYCQ',
                            'IGKV5': 'GVYYCQ',
                            'IGKV6': 'AVYFCQ',
                            'IGKV7': 'THYYCA',
                            'IGKV8': 'AVYYCQ',
                            'IGKV9': 'ADYYCL',
                            'IGLV1': 'AIYFCA',
                            'IGLV2': 'AMYFCA',
                            'IGLV3': 'AIYICG',
                            'IGLV4': 'AIYFCA',
                            'IGLV5': 'AIYFCA',
                            'IGLV6': 'AIYFCA',
                            'IGLV7': 'AIYFCA',
                            'IGLV8': 'AIYFCA'},
            
            'rabbit':    {    'IGHV1': 'ATYFCA',
                            'IGKV1': 'ATYYCQ',
                            'IGLV2': 'AAYYCV',
                            'IGLV3': 'ADYYCQ',
                            'IGLV4': 'ADYYCG',
                            'IGLV5': 'ADYYCA',
                            'IGLV6': 'ADYYCQ'},
            
            'b12mouse':    {    'IGHV1': 'AVYYCA',
                            'IGHV2': 'AIYYCA',
                            'IGHV3': 'ATYYCA',
                            'IGHV4': 'ALYYCA',
                            'IGHV5': 'AMYYCA',
                            'IGHV6': 'GIYYCT',
                            'IGHV7': 'ATYYCA',
                            'IGHV8': 'ATYYCA',
                            'IGHV9': 'ATYFCA',
                            'IGKV1': 'ATYYCQ',
                            'IGKV2': 'GVYYCA',
                            'IGKV3': 'ATYYCQ',
                            'IGKV4': 'ATYYCQ',
                            'IGKV5': 'GVYYCQ',
                            'IGKV6': 'AVYFCQ',
                            'IGKV7': 'THYYCA',
                            'IGKV8': 'AVYYCQ',
                            'IGKV9': 'ADYYCL',
                            'IGLV1': 'AIYFCA',
                            'IGLV2': 'AMYFCA',
                            'IGLV3': 'AIYICG',
                            'IGLV4': 'AIYFCA',
                            'IGLV5': 'AIYFCA',
                            'IGLV6': 'AIYFCA',
                            'IGLV7': 'AIYFCA',
                            'IGLV8': 'AIYFCA'},
            
            'vrc01mouse':{    'IGHV1': 'AVYYCA',
                            'IGHV2': 'AIYYCA',
                            'IGHV3': 'ATYYCA',
                            'IGHV4': 'ALYYCA',
                            'IGHV5': 'AMYYCA',
                            'IGHV6': 'GIYYCT',
                            'IGHV7': 'ATYYCA',
                            'IGHV8': 'ATYYCA',
                            'IGHV9': 'ATYFCA',
                            'IGKV1': 'ATYYCQ',
                            'IGKV2': 'GVYYCA',
                            'IGKV3': 'ATYYCQ',
                            'IGKV4': 'ATYYCQ',
                            'IGKV5': 'GVYYCQ',
                            'IGKV6': 'AVYFCQ',
                            'IGKV7': 'THYYCA',
                            'IGKV8': 'AVYYCQ',
                            'IGKV9': 'ADYYCL',
                            'IGLV1': 'AIYFCA',
                            'IGLV2': 'AMYFCA',
                            'IGLV3': 'AIYICG',
                            'IGLV4': 'AIYFCA',
                            'IGLV5': 'AIYFCA',
                            'IGLV6': 'AIYFCA',
                            'IGLV7': 'AIYFCA',
                            'IGLV8': 'AIYFCA'},

            '9114mouse':{    'IGHV1': 'AVYYCA',
                            'IGHV2': 'AIYYCA',
                            'IGHV3': 'ATYYCA',
                            'IGHV4': 'ALYYCA',
                            'IGHV5': 'AMYYCA',
                            'IGHV6': 'GIYYCT',
                            'IGHV7': 'ATYYCA',
                            'IGHV8': 'ATYYCA',
                            'IGHV9': 'ATYFCA',
                            'IGKV1': 'ATYYCQ',
                            'IGKV2': 'GVYYCA',
                            'IGKV3': 'ATYYCQ',
                            'IGKV4': 'ATYYCQ',
                            'IGKV5': 'GVYYCQ',
                            'IGKV6': 'AVYFCQ',
                            'IGKV7': 'THYYCA',
                            'IGKV8': 'AVYYCQ',
                            'IGKV9': 'ADYYCL',
                            'IGLV1': 'AIYFCA',
                            'IGLV2': 'AMYFCA',
                            'IGLV3': 'AIYICG',
                            'IGLV4': 'AIYFCA',
                            'IGLV5': 'AIYFCA',
                            'IGLV6': 'AIYFCA',
                            'IGLV7': 'AIYFCA',
                            'IGLV8': 'AIYFCA'}
            }

        return germs[vdj.species][trunc_gene]


    def _get_j_chunk(self, vdj):
        trunc_gene = vdj.j.top_germline[:5]
        germs = {
            'human':    {    'IGHJ1': 'WGQGT',
                            'IGHJ2': 'WGRGT',
                            'IGHJ3': 'WGQGT',
                            'IGHJ4': 'WGQGT',
                            'IGHJ5': 'WGQGT',
                            'IGHJ6': 'WGQGT',
                            'IGKJ1': 'FGQGT',
                            'IGKJ2': 'FGQGT',
                            'IGKJ3': 'FGPGT',
                            'IGKJ4': 'FGGGT',
                            'IGKJ5': 'FGQGT',
                            'IGLJ1': 'FGTGT',
                            'IGLJ2': 'FGGGT',
                            'IGLJ3': 'FGGGT',
                            'IGLJ4': 'FGGGT',
                            'IGLJ5': 'FGEGT',
                            'IGLJ6': 'FGSGT',
                            'IGLJ7': 'FGGGT'},

            'macaque':    {    'IGHJ1': 'WGQGA',
                            'IGHJ2': 'WGPGT',
                            'IGHJ3': 'WGQGL',
                            'IGHJ4': 'WGQGV',
                            'IGHJ5': 'WGPGV',
                            'IGHJ6': 'WGQGV',
                            'IGKJ1': 'FGQGT',
                            'IGKJ2': 'FGQGT',
                            'IGKJ3': 'FGPGT',
                            'IGKJ4': 'FGGGT',
                            'IGKJ5': 'FGQGT',
                            'IGLJ1': 'FGAGT',
                            'IGLJ2': 'FGGGT',
                            'IGLJ3': 'FGGGT',
                            'IGLJ4': 'FCGGT',
                            'IGLJ5': 'FGEGT',
                            'IGLJ6': 'FGSGT'},

            'mouse':    {    'IGHJ1': 'WGAGT',
                            'IGHJ2': 'WGQGT',
                            'IGHJ3': 'WGQGT',
                            'IGHJ4': 'WGQGT',
                            'IGKJ1': 'FGGGT',
                            'IGKJ2': 'FGGGT',
                            'IGKJ4': 'FGSGT',
                            'IGKJ5': 'FGAGT',
                            'IGLJ1': 'FGGGT',
                            'IGLJ2': 'FGGGT',
                            'IGLJ3': 'FGSGT'},

            'rabbit':    {    'IGHJ1': 'WGTGT',
                            'IGHJ2': 'WGPGT',
                            'IGHJ3': 'WGQGT',
                            'IGHJ4': 'WGPGT',
                            'IGHJ5': 'WGQGT',
                            'IGHJ6': 'WGPGT',
                            'IGKJ1': 'FGGGT',
                            'IGKJ1': 'FGAGT',
                            'IGLJ5': 'FGGGT',
                            'IGLJ6': 'FGGGT'},

            'b12mouse':    {    'IGHJ1': 'WGAGT',
                            'IGHJ2': 'WGQGT',
                            'IGHJ3': 'WGQGT',
                            'IGHJ4': 'WGQGT',
                            'IGHJ6': 'WGQGT',
                            'IGKJ1': 'FGGGT',
                            'IGKJ2': 'FGGGT',
                            'IGKJ4': 'FGSGT',
                            'IGKJ5': 'FGAGT',
                            'IGLJ1': 'FGGGT',
                            'IGLJ2': 'FGGGT',
                            'IGLJ3': 'FGSGT'},

            'vrc01mouse':{    'IGHJ1': 'WGAGT',
                            'IGHJ2': 'WGQGT',
                            'IGHJ3': 'WGQGT',
                            'IGHJ4': 'WGQGT',
                            'IGHJ6': 'WGQGT',
                            'IGKJ1': 'FGGGT',
                            'IGKJ2': 'FGGGT',
                            'IGKJ4': 'FGSGT',
                            'IGKJ5': 'FGAGT',
                            'IGLJ1': 'FGGGT',
                            'IGLJ2': 'FGGGT',
                            'IGLJ3': 'FGSGT'},

            '9114mouse':{    'IGHJ1': 'WGAGT',
                            'IGHJ2': 'WGQGT',
                            'IGHJ3': 'WGQGT',
                            'IGHJ4': 'WGQGT',
                            'IGHJ6': 'WGQGT',
                            'IGKJ1': 'FGGGT',
                            'IGKJ2': 'FGGGT',
                            'IGKJ4': 'FGSGT',
                            'IGKJ5': 'FGAGT',
                            'IGLJ1': 'FGGGT',
                            'IGLJ2': 'FGGGT',
                            'IGLJ3': 'FGSGT'}
            }

        return germs[vdj.species][trunc_gene]



#!/usr/bin/python
# filename: pandaseq.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import sys
import glob
import subprocess as sp
from multiprocessing import cpu_count


def list_files(d):
    return sorted([f for f in glob.glob(d + '/*') if os.path.isfile(f)])


def pair_files(files, nextseq):
    pairs = {}
    for f in files:
        if nextseq:
            f_prefix = '_'.join(os.path.basename(f).split('_')[:3])
        else:
            f_prefix = '_'.join(os.path.basename(f).split('_')[:2])
        if f_prefix in pairs:
            pairs[f_prefix].append(f)
        else:
            pairs[f_prefix] = [f, ]
    return pairs


def batch_pandaseq(f, r, o, algo):
    cmd = 'pandaseq -f {0} -r {1} -A {2} -d rbfkms -T {3} -w {4}'.format(f, r, algo, cpu_count(), o)
    sp.Popen(cmd, shell=True, stderr=sp.STDOUT, stdout=sp.PIPE).communicate()


def merge_reads(files, output, args, i):
    files.sort()
    f = files[0]
    r = files[1]
    if args.nextseq:
        lane = os.path.basename(f).split('_')[-3]
        sample_id = os.path.basename(f).split('_')[0]
        sample = sample_id + '_' + lane
    else:
        sample = os.path.basename(f).split('_')[0]
    print_sample_info(i, sample)
    o = os.path.join(output, '{}.fasta'.format(sample))
    batch_pandaseq(f, r, o, args.pandaseq_algo)
    # print_sample_end(log)


def print_start_info():
    sys.stdout.write('\n')
    sys.stdout.write('\n')
    sys.stdout.write('========================================\n')
    sys.stdout.write('Merging reads with PANDAseq\n')
    sys.stdout.write('========================================\n')
    sys.stdout.write('\n')


def print_input_info(files):
    sys.stdout.write('The input directory contains {} pair(s) of files to be merged.\n\n'.format(len(files) / 2))


def print_sample_info(i, sample):
    sys.stdout.write('[ {} ]  Processing sample {}\n'.format(str(i), sample))


def print_sample_end():
    sys.stdout.write('Done.\n')


def run(input, output, args, log=False):
    if log:
        import logging
    try:
        nextseq = args.nextseq
    except:
        args.nextseq = False
    print_start_info()
    files = list_files(input)
    print_input_info(files)
    pairs = pair_files(files, args.nextseq)
    for i, pair in enumerate(sorted(pairs.keys())):
        if len(pairs[pair]) == 2:
            merge_reads(pairs[pair], output, args, i)
            if log:
                file1 = os.path.basename(pairs[pair][0])
                file2 = os.path.basename(pairs[pair][1])
                logging.info('Merging {} and {}'.format(file1, file2))



from setuptools import setup

config = {
	'description': 'VDJ assignment and antibody sequence annotation. Scalable from a single sequence to billions of sequences.',
	'author': 'Bryan Briney',
	'url': 'www.github.com/briney/abstar/',
	# 'download_url': 'www.github.com/briney/abstar/',
	'author_email': 'briney@scripps.edu',
	'version': '0.1.2',
	'install_requires': ['abtools',
						 'biopython',
						 'celery',
						 'nwalign',
						 'pymongo',
						 'scikit-bio'],
	'packages': ['abstar'],
	'scripts': ['bin/abstar',
				'bin/batch_mongoimport',
				'bin/make_basespace_credfile'],
	'name': 'abstar',
	'include_package_data': True
}

setup(**config)



#!/usr/bin/python
# filename: blast.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import os
import platform
import tempfile

from Bio import SeqIO
from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.Blast import NCBIXML

from abtools.alignment import local_alignment


class BlastResult(object):
    '''
    Data structure for parsing and holding a BLASTn result.
    Input is a file handle for the XML-formatted BLASTn output file.
    '''
    def __init__(self, seq, blastout, species):
        super(BlastResult, self).__init__()
        self.seq = seq
        self.id = seq.id
        self.alignments = blastout.alignments
        self.input_sequence = seq.sequence
        self.species = species
        self.top_germline = self._get_top_germline()
        self.all_germlines = self._get_all_germlines()
        self.top_score = self._get_top_score()
        self.all_scores = self._get_all_scores()
        self.top_evalue = self._get_top_evalue()
        self.all_evalues = self._get_all_evalues()
        self.top_bitscore = self._get_top_bitscore()
        self.all_bitscores = self._get_all_bitscores()
        self.strand = 'plus'
        self.query_alignment = self._get_query_alignment()
        self.germline_alignment = self._get_germline_alignment()
        self.alignment_midline = self._get_alignment_midline()
        self.alignment_length = self._get_alignment_length()
        self.query_start = self._get_query_start()
        self.query_end = self._get_query_end()
        self.germline_start = self._get_germline_start()
        self.germline_end = self._get_germline_end()
        self.gene_type = self._gene_type()
        self.chain = self._chain()

    def annotate(self):
        self.fs_indel_adjustment = 0
        self.nfs_indel_adjustment = 0
        self._fix_ambigs()
        self._find_indels()
        self.regions = self._regions()
        self.nt_mutations = self._nt_mutations()
        self.aa_mutations = self._aa_mutations()


    def realign_variable(self, germline_gene):
        '''
        Due to restrictions on the available scoring parameters in BLASTn, incorrect truncation
        of the v-gene alignment can occur. This function re-aligns the query sequence with
        the identified germline variable gene using more appropriate alignment parameters.

        Input is the name of the germline variable gene (ex: 'IGHV1-2*02').
        '''
        self.germline_seq = self._get_germline_sequence_for_realignment(germline_gene, 'V')
        alignment = local_alignment(self.seq.sequence, self.germline_seq,
                                    gap_open_penalty=22, gap_extend_penalty=1)
        rc = self.seq.reverse_complement
        alignment_rc = local_alignment(rc, self.germline_seq,
                                       gap_open_penalty=22, gap_extend_penalty=1)
        if alignment.score > alignment_rc.score:
            self._process_realignment(alignment)
        else:
            self.strand = 'minus'
            self.input_sequence = rc
            self._process_realignment(alignment_rc)


    def _get_germline_sequence_for_realignment(self, germ, gene):
        '''
        Identifies the appropriate germline variable gene from a database of all
        germline variable genes.

        Input is the name of the germline variable gene (ex: 'IGHV1-2*02') and
        the gene region ('V' or 'J').

        Output is the germline sequence.
        '''
        mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        db_file = os.path.join(mod_dir, 'ssw/dbs/{}_{}.fasta'.format(self.species.lower(), gene))
        for s in SeqIO.parse(open(db_file), 'fasta'):
            if s.id == germ:
                return str(s.seq)
        return None


    def _process_realignment(self, alignment):
        '''
        Processes the result of variable gene realignment and updates BlastResult
        attributes accordingly.

        Input is an AbTools SSWAlignment object.
        '''
        self.query_alignment = alignment.aligned_query
        self.germline_alignment = alignment.aligned_target
        self.alignment_midline = alignment.alignment_midline
        if self.gene_type == 'variable':
            self.query_start = alignment.query_begin
            self.germline_start = alignment.target_begin
        self.query_end = alignment.query_end
        self.germline_end = alignment.target_end


    def _fix_ambigs(self):
        '''
        Fixes ambiguous nucleotides by replacing them with the germline nucleotide.
        '''
        from abstar.utils import ambigs
        self.query_alignment = ambigs.fix_ambigs(self)


    def _find_indels(self):
        '''
        Identifies and annotates indels in the query sequence.
        '''
        self.insertions = []
        self.deletions = []
        if self._indel_check():
            from abstar.utils import indels
            self.insertions = indels.find_insertions(self)
            self.deletions = indels.find_deletions(self)

    def _indel_check(self):
        '''
        Checks the sequence for evidence of insertions or deletions.
        '''
        if '-' in self.query_alignment:
            return True
        elif '-' in self.germline_alignment:
            return True
        return False

    def _regions(self):
        '''
        Identifies and annotates variable/joining gene regions.
        '''
        from abstar.utils import regions
        return regions.regions(self)

    def _nt_mutations(self):
        '''
        Identifies and annotates nucleotide mutations.
        '''
        from abstar.utils import mutations
        return mutations.nt_mutations(self)

    def _aa_mutations(self):
        '''
        Identifies and annotates amino acid mutations.
        '''
        from abstar.utils import mutations
        return mutations.aa_mutations(self)

    def _get_top_germline(self):
        'Returns the top scoring germline gene'
        return self.alignments[0].title.split()[0]

    def _get_all_germlines(self):
        'Returns all germline genes'
        return [alignment.title.split()[0] for alignment in self.alignments]

    def _get_top_score(self):
        'Returns the score for the top scoring germline gene'
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].score

    def _get_all_scores(self):
        'Returns all germline gene scores'
        return [alignment.hsps[0].score for alignment in self.alignments]

    def _get_top_evalue(self):
        'Returns the e-value for the top scoring germline gene'
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].expect

    def _get_all_evalues(self):
        return [alignment.hsps[0].expect for alignment in self.alignments]

    def _get_top_bitscore(self):
        'Returns the bitscore for the top scoring germline gene'
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].bits

    def _get_all_bitscores(self):
        return [alignment.hsps[0].bits for alignment in self.alignments]

    def _get_strand(self):
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].strand

    def _get_query_alignment(self):
        '''Returns the query alignment string for the
        top scoring germline gene alignment'''
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].query

    def _get_germline_alignment(self):
        'Returns the top scoring germline alignment'
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].sbjct

    def _get_alignment_midline(self):
        '''Returns the alignment midline string for the
        top scoring germline gene alignment'''
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].match

    def _get_alignment_length(self):
        'Returns the alignment length for the top scoring germline gene alignment'
        return self.alignments[0].length

    def _get_query_start(self):
        '''Returns the start position of the query alignment with
        the top germline gene'''
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].query_start - 1

    def _get_query_end(self):
        '''Returns the start position of the query alignment with
        the top germline gene'''
        return self.query_start + self.alignment_length

    def _get_germline_start(self):
        'Returns the start position of the top germline alignment'
        top_alignment = self.alignments[0]
        return top_alignment.hsps[0].sbjct_start - 1

    def _get_germline_end(self):
        'Returns the end position of the top germline alignment'
        return self.germline_start + self.alignment_length

    def _gene_type(self):
        "Returns the gene type, either 'variable' or 'joining'"
        if self.top_germline[3] == 'V':
            return 'variable'
        elif self.top_germline[3] == 'J':
            return 'joining'
        return None

    def _chain(self):
        'Returns the chain type'
        if self.top_germline.startswith('IGH'):
            return 'heavy'
        elif self.top_germline.startswith('IGK'):
            return 'kappa'
        elif self.top_germline.startswith('IGL'):
            return 'lambda'
        # hack to accomodate IgBLAST's crappy germline database
        # to be removed when I curate the database
        elif self.top_germline.startswith('VH'):
            return 'heavy'
        return None


class DiversityResult(object):
    """
    Data structure for holding information about diversity germline gene assignments.
    Designed to have (mostly) the same attributes as BlastResult objects, so that
    DiversityResult objects and BlastResult objects can (mostly) be used
    interchangeably in downstream operations.

    Main differences between DiversityResult and BlastResult are that DiversityResults
    have empty bitscore and e-value attributes, and the alignments attribute contains
    an Alignment object instead of parsed BLASTn output.

    Input is a list of Alignment objects, representing the top-scoring diversity genes.
    """
    def __init__(self, seq, alignments):
        super(DiversityResult, self).__init__()
        self.id = seq.id
        self.input_sequence = seq.sequence
        self.alignments = alignments
        self.top_germline = self._get_top_germline()
        self.all_germlines = self._get_all_germlines()
        self.top_score = self._get_top_score()
        self.all_scores = self._get_all_scores()
        self.top_evalue = None
        self.all_evalues = []
        self.top_bitscore = None
        self.all_bitscores = []
        self.query_alignment = self._get_query_alignment()
        self.germline_alignment = self._get_germline_alignment()
        self.alignment_midline = self._get_alignment_midline()
        self.alignment_length = self._get_alignment_length()
        self.query_start = self._get_query_start()
        self.query_end = self._get_query_end()
        self.germline_start = self._get_germline_start()
        self.germline_end = self._get_germline_end()
        self.sequence = self._get_sequence()
        self.reading_frame = self._get_reading_frame()
        self.gene_type = 'diversity'
        self.chain = 'heavy'
        self.nt_mutations = self._nt_mutations()

    def _get_top_germline(self):
        '''
        Returns the top scoring germline gene. If no germline gene scores
        higher than 9, then the top_germline attribute is set to 'None'.
        The minimum for reaching a score of 9 would be a 6nt alignment
        region with at least 5 matching nucleotides and a single mismatch.
        '''
        top_alignment = self.alignments[0]
        if top_alignment.score > 9:
            return top_alignment.target.id
        return None

    def _get_all_germlines(self):
        'Returns all germline genes'
        return [a.target.id for a in self.alignments]

    def _get_top_score(self):
        'Returns the score for the top scoring germline gene'
        top_alignment = self.alignments[0]
        return top_alignment.score

    def _get_all_scores(self):
        'Returns all germline gene scores'
        return [a.score for a in self.alignments]

    def _get_query_alignment(self):
        '''Returns the query alignment string for the
        top scoring germline gene alignment'''
        top_alignment = self.alignments[0]
        return top_alignment.aligned_query

    def _get_germline_alignment(self):
        'Returns the top scoring germline alignment'
        top_alignment = self.alignments[0]
        return top_alignment.aligned_target

    def _get_alignment_midline(self):
        '''Returns the alignment midline string for the
        top scoring germline gene alignment'''
        query = self.query_alignment
        germ = self.germline_alignment
        midline = ['|' if q == g else ' ' for q, g in zip(query, germ)]
        return ''.join(midline)

    def _get_alignment_length(self):
        'Returns the alignment length for the top scoring germline gene alignment'
        return len(self.germline_alignment)

    def _get_query_start(self):
        '''Returns the start position of the query alignment with
        the top germline gene'''
        top_alignment = self.alignments[0]
        return top_alignment.query_begin

    def _get_query_end(self):
        '''Returns the start position of the query alignment with
        the top germline gene'''
        top_alignment = self.alignments[0]
        return top_alignment.query_end

    def _get_germline_start(self):
        'Returns the start position of the top germline alignment'
        top_alignment = self.alignments[0]
        return top_alignment.target_begin

    def _get_germline_end(self):
        'Returns the end position of the top germline alignment'
        top_alignment = self.alignments[0]
        return top_alignment.target_end

    def _get_reading_frame(self):
        '''
        Identifies the diverstiy gene reading frame
        '''
        rf = (self.germline_start % 3) + 1
        return rf

    def _get_sequence(self):
        return self.query_alignment[self.query_start:self.query_end]

    def _nt_mutations(self):
        '''
        Identifies and annotates nucleotide mutations.
        '''
        from abstar.utils import mutations
        return mutations.nt_mutations(self)


def build_j_blast_input(seqs, v_blast_results):
    j_fastas = []
    for seq, vbr in zip(seqs, v_blast_results):
        start = len(vbr.query_alignment) + vbr.query_start + vbr.fs_indel_adjustment + vbr.nfs_indel_adjustment
        j_fastas.append('>{}\n{}'.format(seq.id, seq.sequence[start:]))
    j_blastin = tempfile.NamedTemporaryFile(delete=False)
    j_blastin.write('\n'.join(j_fastas))
    j_blastin.close()
    return j_blastin


def blast(seq_file, species, segment):
    '''
    Runs BLASTn against an antibody germline database.

    Input is a FASTA file of sequences (the file path, not a handle), the species of origin
    of the sequences to be queried, and the gene segment (options are: 'V', 'D', or 'J')
    '''
    mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    blast_path = os.path.join(mod_dir, 'blast/blastn_{}'.format(platform.system().lower()))
    blast_db_path = os.path.join(mod_dir, 'blast/dbs/{}_gl_{}'.format(species.lower(), segment.upper()))
    blastout = tempfile.NamedTemporaryFile(delete=False)
    blastn_cmd = NcbiblastnCommandline(cmd=blast_path,
                                       db=blast_db_path,
                                       query=seq_file,
                                       out=blastout.name,
                                       outfmt=5,
                                       dust='no',
                                       word_size=_word_size(segment),
                                       max_target_seqs=10,
                                       evalue=_evalue(segment),
                                       reward=_match_reward(segment),
                                       penalty=_mismatch_penalty(segment),
                                       gapopen=_gap_open(segment),
                                       gapextend=_gap_extend(segment))
    stdout, stderr = blastn_cmd()
    blast_records = [br for br in NCBIXML.parse(blastout)]
    os.unlink(blastout.name)
    return blast_records


def _word_size(segment):
    'Returns BLASTn word size for the given gene segment'
    word_sizes = {'V': 11,
                  'D': 4,
                  'J': 7}
    return word_sizes[segment]


def _gap_open(segment):
    'Returns BLASTn gap-open penalty for the given gene segment'
    gap_open = {'V': 5,
                'D': 4,
                'J': 5}
    return gap_open[segment]


def _gap_extend(segment):
    'Returns BLASTn gap-extend penalty for the given gene segment'
    gap_extend = {'V': 2,
                  'D': 2,
                  'J': 2}
    return gap_extend[segment]


def _match_reward(segment):
    'Returns BLASTn match reward for the given gene segment'
    match = {'V': 1,
             'D': 1,
             'J': 1}
    return match[segment]


def _mismatch_penalty(segment):
    'Returns BLASTn mismatch penalty for the given gene segment'
    mismatch = {'V': -1,
                'D': -1,
                'J': -1}
    return mismatch[segment]


def _evalue(segment):
    'Returns minimum BLASTn e-value for the given gene segment'
    evalue = {'V': 1,
              'D': 100000,
              'J': 1000}
    return evalue[segment]
